/*-
 * Copyright (c) 2010 Andrey Zonov <andrey.zonov@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/mman.h>
#include <sys/param.h>
#include <sys/proc.h>
#include <sys/stat.h>
#include <sys/sysctl.h>
#include <sys/user.h>

#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <kvm.h>
#include <paths.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>

#define MIN_PID	48		/* sys/vm/vm_pageout.c */
#define MAX_PID	99999		/* <sys/proc.h> */
#define MIN_TIMEOUT	1	/* in seconds */
#define MAX_TIMEOUT	600	/* in seconds */
#define TIMEOUT MIN_TIMEOUT	/* in seconds */
#define MIN_DEBUG_LEVEL	0	/* quiet */
#define MAX_DEBUG_LEVEL	3	/* debug */
#define DEBUG_LEVEL	1	/* only errors */

static int fd;
static int kq;
static int numpids;
static int syslog_flag;
static int debug_level;
static int daemonize_flag;
static int logfd;
static char *logfile;
static char **pidfiles;
static pid_t *ppids;
static unsigned int timeout;
static void *ptr;
static kvm_t *kd;

static void usage(void);
static void init(void);
static void loop(void);
static int set_protect(pid_t ppid);
static pid_t takepid(const char *pidfile);
static void cleanup(void);
static void loginit(void);
static void logout(const int level, const char *fmt, ...);
static void logerr(const int level, const char *fmt, ...);

int
main(int argc, char **argv)
{
	int i;
	int ch;
	int len;
	int arg;
	char *endptr;
	char *listpids;

	debug_level = DEBUG_LEVEL;
	logfile = "syslog";
	syslog_flag = 1;
	daemonize_flag = 1;
	numpids = 0;
	timeout = TIMEOUT;
	arg = argc;

	while ((ch = getopt(argc, argv, "d:F:l:nt:")) != -1) {
		switch (ch) {
		case 'd':
			debug_level = strtol(optarg, &endptr, 10);
			if (*optarg == '\0' || *endptr != '\0') {
				fprintf(stderr, "Invalid debug level: %s\n", optarg);
				usage();
			}
			if (debug_level < MIN_DEBUG_LEVEL || debug_level > MAX_DEBUG_LEVEL) {
				fprintf(stderr, "Debug level must be %d..%d\n", MIN_DEBUG_LEVEL, MAX_DEBUG_LEVEL);
				usage();
			}

			arg -= 2;

			break;
		case 'F':

			for (numpids = 0, len = 0, listpids = optarg; *listpids; ++listpids) {
				if (*listpids == ',') {
					if (len == 0) {
						fprintf(stderr, "Not allowed empty pidfile\n");
						usage();
					}
					numpids++;
					len = 0;
				} else 
					len++;
			}
			if (len == 0) {
				fprintf(stderr, "Not allowed empty pidfile\n");
				usage();
			}
			numpids++;

			pidfiles = (char **) malloc(sizeof(char *) * numpids);
			if (pidfiles == NULL) {
				fprintf(stderr, "Cannot allocate memory");
				exit(1);
			}

			for (i = 0, len = 0, listpids = optarg; *listpids; ++listpids) {
				if (*listpids == ',') {
					pidfiles[i] = (char *) malloc(len + 1);
					if (pidfiles[i] == NULL) {
						fprintf(stderr, "Cannot allocate memory");
						exit(1);
					}
					memcpy(pidfiles[i], listpids - len, len);
					pidfiles[i][len] = '\0';
					i++;
					len = 0;
				} else
					len++;
			}
			pidfiles[i] = (char *) malloc(len + 1);
			if (pidfiles[i] == NULL) {
				fprintf(stderr, "Cannot allocate memory");
				exit(1);
			}
			memcpy(pidfiles[i], listpids - len, len);
			pidfiles[i][len] = '\0';

			arg -= 2;

			break;
		case 'l':
			logfile = optarg;
			if (!strcmp(logfile, "syslog"))
				syslog_flag = 1;
			else
				syslog_flag = 0;

			arg -= 2;

			break;
		case 'n':
			daemonize_flag = 0;

			arg--;

			break;
		case 't':
			timeout = strtol(optarg, &endptr, 10);
			if (*optarg == '\0' || *endptr != '\0') {
				fprintf(stderr, "Invalid timeout: %s\n", optarg);
				usage();
			}
			if (timeout < MIN_TIMEOUT || timeout > MAX_TIMEOUT) {
				fprintf(stderr, "Timeout must be %d..%d\n", MIN_TIMEOUT, MAX_TIMEOUT);
				usage();
			}

			arg -= 2;

			break;
		default:
			usage();
			break;
		}
	}

	if (arg != 1)
		usage();

	if (numpids == 0)
		usage();

	init();
	loop();
	cleanup();

	exit(0);
}

static void
usage(void)
{
	fprintf(stderr, "usage: %s [-n] [-d debug level] [-l logfile] [-t timeout]"
	    " -F pidfile1,pidfile2,...\n", getprogname());
	exit(1);
}

static void
init(void)
{
	int rc;
	
	kd = kvm_open(NULL, _PATH_MEM, NULL, O_RDONLY, NULL);
	if (kd == NULL)
		err(1, "kvm_open(%s)", _PATH_MEM);

	fd = open(_PATH_KMEM, O_RDWR, 0);
	if (fd == -1)
		err(1, "open(%s)", _PATH_KMEM);

	ppids = (pid_t *) malloc(sizeof(pid_t) * numpids);
	if (ppids == NULL)
		err(1, "malloc()");

	loginit();

	if (daemonize_flag)
		if (daemon(0, 0) == -1)
			err(1, "Cannot daemonize");

	rc = madvise(NULL, 0, MADV_PROTECT);
	if (rc == -1) {
		logerr(1, "madvise()");
		exit(1);
	}

	kq = kqueue();
	if (kq == -1) {
		logerr(1, "kqueue()");
		exit(1);
	}
}

static void
loop(void)
{
	int i;
	int n;
	int nc;
	int need_sleep;
	struct kevent *ch;

	nc = 0;

	ch = (struct kevent *) calloc(1, sizeof(struct kevent) * numpids);
	if (ch == NULL) {
		logerr(1, "malloc()");
		exit(1);
	}

	for (i = 0; i < numpids; i++) {
		ppids[i] = takepid(pidfiles[i]);
		if (ppids[i] != -1)
			if (set_protect(ppids[i]) == -1)
				ppids[i] = -1;
	}
	
	for( ;; ) {
		for (i = 0; i < numpids; i++)
			EV_SET(&ch[nc++], ppids[i], EVFILT_PROC, EV_ADD | EV_ENABLE | EV_ONESHOT, NOTE_EXIT, 0, NULL);

		n = kevent(kq, ch, nc, ch, numpids, NULL);
		if (n == -1)
			logerr(1, "kevent()");

		nc = 0;
		need_sleep = 0;
		for (i = 0; i < numpids; i++) {
			ppids[i] = takepid(pidfiles[i]);
			if (ppids[i] == -1)
				need_sleep = 1;
			else
				if (set_protect(ppids[i]) == -1) {
					ppids[i] = -1;
					need_sleep = 1;
				}
		}

		if (need_sleep)
			sleep(timeout);
	}
}

static int
set_protect(pid_t ppid)
{
	int rc;
	int cnt;
	off_t off;
	struct kinfo_proc *kip;
	struct proc *p_mmap;

	if (ppid <= 0)
		return(-1);

	kip = kvm_getprocs(kd, KERN_PROC_PID, ppid, &cnt);
	if (kip == NULL || cnt != 1) {
		logerr(2, "Cannot get kinfo for pid %d", ppid);
		return(-1);
	}

	logout(3, "Flags for pid %d: %lX", ppid, kip->ki_flag);
	logout(3, "Address for pid %d: %p", ppid, (void *)kip->ki_paddr);

	if (kip->ki_flag & P_PROTECTED) {
		logout(3, "P_PROTECTED flag is already set for pid %d", ppid);
		return(0);
	}

	off = (off_t)((uintptr_t)kip->ki_paddr);
	ptr = mmap(0, sizeof(struct proc), PROT_READ | PROT_WRITE, MAP_SHARED, fd, off);
	if (ptr == MAP_FAILED) {
		logerr(1, "mmap()");
		return(-1);
	}

	p_mmap = (struct proc *)ptr;
	p_mmap->p_flag |= P_PROTECTED;

	rc = munmap(ptr, sizeof(struct proc));
	if (rc == -1) {
		logerr(1, "munmap()");
		return(-1);
	}

	/* FIXME: flags not yet updated here */
	logout(3, "Flags for pid %d: %lX", ppid, kip->ki_flag);

	return(0);
}

static pid_t
takepid(const char *pidfile)
{
	char buf[BUFSIZ];
	char *endp;
	pid_t ppid;
	FILE *ppidfh;

	ppidfh = NULL;

	ppidfh = fopen(pidfile, "r");
	if (ppidfh == NULL) {
		logerr(2, "Cannot open pidfile: %s", pidfile);
		return(-1);
	}

	if (fgets(buf, BUFSIZ, ppidfh) == NULL) {
		if (feof(ppidfh))
			logout(2, "Empty pidfile: %s", pidfile);
		else
			logout(2, "Cannot read pidfile: %s", pidfile);
		fclose(ppidfh);
		return(-1);
	}

	fclose(ppidfh);

	ppid = strtol(buf, &endp, 10);
	if (*endp != '\0' && !isspace((unsigned char)*endp)) {
		logout(2, "Invalid pidfile: %s", pidfile);
		return(-1);
	} else if (ppid < MIN_PID || ppid > MAX_PID) {
		logout(2, "Invalid range pid: %d", ppid);
		return(-1);
	}

	return(ppid);
}

static void
cleanup(void)
{
	closelog();
	close(fd);
	close(kq);
	kvm_close(kd);
}

static void
loginit(void)
{
	if (debug_level == MIN_DEBUG_LEVEL)
		return;

	if (syslog_flag)
		openlog(getprogname(), LOG_CONS | LOG_PID, LOG_DAEMON);
	else {
		logfd = open(logfile, O_RDWR | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
		if (logfd == -1)
			err(1, "Cannot open logfile: %s", logfile);
	}
}

static void
logout(const int level, const char *fmt, ...)
{
	int n, m;
	long int usec;
	char *nlevel;
	char buf[PIPE_BUF], *bufp;
	va_list ap;
	time_t tv_time_t;
	struct timeval tv;
	struct tm *lt;

	n = m = 0;
	bufp = buf;

	if (level > debug_level)
		return;

	if (syslog_flag) {
		va_start(ap, fmt);
		n = vsnprintf(bufp, PIPE_BUF, fmt, ap);
		va_end(ap);

		switch(level) {
		case 1:
			syslog(LOG_ERR, buf);
			break;
		case 2:
			syslog(LOG_WARNING, buf);
			break;
		case 3:
			syslog(LOG_DEBUG, buf);
			break;
		default:
			break;
		}

		return;
	}

	gettimeofday(&tv, (struct timezone *)NULL);
	tv_time_t = tv.tv_sec;
	lt = localtime(&tv_time_t);
	lt->tm_mon++;
	lt->tm_year += 1900;
	usec = tv.tv_usec;

	switch(level) {
		case 1:
			nlevel = "error";
			break;
		case 2:
			nlevel = "warn";
			break;
		case 3:
			nlevel = "debug";
			break;
		default:
			break;
	}

	n = snprintf(bufp, PIPE_BUF, "%02d/%02d/%04d %02d:%02d:%02d.%06ld [%s] ",
	    lt->tm_mday, lt->tm_mon, lt->tm_year, lt->tm_hour, lt->tm_min, lt->tm_sec, usec, nlevel);
	bufp += n;
	m +=n;

	va_start(ap, fmt);
	if (fmt != NULL) {
		n = vsnprintf(bufp, (PIPE_BUF - m), fmt, ap);
		bufp += n;
		m += n;
	}
	va_end(ap);

	if (m >= PIPE_BUF) {
		bufp += - m + PIPE_BUF - 2;
		m = PIPE_BUF - 2;
	}

	n = snprintf(bufp, (PIPE_BUF - m), "\n");
	m += n;

	write(logfd, buf, m);
}

static void
logerr(const int level, const char *fmt, ...)
{
	int n, m;
	long int usec;
	char *nlevel;
	char buf[PIPE_BUF], *bufp, errbuf[PIPE_BUF], *errbufp;
	va_list ap;
	time_t tv_time_t;
	struct timeval tv;
	struct tm *lt;

	n = m = 0;
	bufp = buf;
	errbufp = errbuf;

	if (level > debug_level)
		return;

	if (syslog_flag) {
		va_start(ap, fmt);
		n = vsnprintf(bufp, PIPE_BUF, fmt, ap);
		snprintf(bufp + n, (PIPE_BUF - n), " (%d: %%m)", errno);
		va_end(ap);

		switch(level) {
		case 1:
			syslog(LOG_ERR, buf);
			break;
		case 2:
			syslog(LOG_WARNING, buf);
			break;
		case 3:
			syslog(LOG_DEBUG, buf);
			break;
		default:
			break;
		}

		return;
	}

	gettimeofday(&tv, (struct timezone *)NULL);
	tv_time_t = tv.tv_sec;
	lt = localtime(&tv_time_t);
	lt->tm_mon++;
	lt->tm_year += 1900;
	usec = tv.tv_usec;

	switch(level) {
		case 1:
			nlevel = "error";
			break;
		case 2:
			nlevel = "warn";
			break;
		case 3:
			nlevel = "debug";
			break;
		default:
			break;
	}

	n = snprintf(bufp, PIPE_BUF, "%02d/%02d/%04d %02d:%02d:%02d.%06ld [%s] ",
	    lt->tm_mday, lt->tm_mon, lt->tm_year, lt->tm_hour, lt->tm_min, lt->tm_sec, usec, nlevel);
	bufp += n;
	m +=n;

	va_start(ap, fmt);
	if (fmt != NULL) {
		n = vsnprintf(bufp, (PIPE_BUF - m), fmt, ap);
		bufp += n;
		m += n;
	}
	va_end(ap);

	strerror_r(errno, errbufp, PIPE_BUF - m);
	n = snprintf(bufp, (PIPE_BUF - m), " (%d: %s)", errno, errbuf);
	bufp += n;
	m += n;

	if (m >= PIPE_BUF) {
		bufp += - m + PIPE_BUF - 2;
		m = PIPE_BUF - 2;
	}

	n = snprintf(bufp, (PIPE_BUF - m), "\n");
	m += n;

	write(logfd, buf, m);
}
